<?php

use CleverReach\Infrastructure\Interfaces\Required\Configuration;

/**
 * Configuration service.
 *
 * @see \CleverReach\Infrastructure\Interfaces\Required\Configuration
 */
class CleverReachConfigService implements Configuration {
  const CLIENT_ID = 'nRiRL1I0WF';
  const CLIENT_SECRET = 'tdOshpUD6LvSq9h0RiXklkcDcYFUYCAr';

  const MODULE_NAME = 'cleverreach';
  const INTEGRATION_NAME = 'Drupal';
  const SUBSCRIPTION_FIELD = 'field_cleverreach_subscribed';

  const CLEVERREACH_HELP_URL = 'https://support.cleverreach.de/hc/en-us/requests/new';
  const CLEVERREACH_BUILD_EMAIL_URL = '/admin/login.php?ref=%2Fadmin%2Fmailing_create_new.php';

  /**
   * CleveReach API access token.
   *
   * @var string
   */
  private $accessToken;
  /**
   * User info data.
   *
   * @var array
   */
  private $userInfo;
  /**
   * Context data, not used in this integration.
   *
   * @var string
   */
  private $context = '';

  /**
   * Saves min log level in integration database.
   *
   * @param int $minLogLevel
   *   Defined log level.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function saveMinLogLevel($minLogLevel) {
    $this->set('min_log_level', $minLogLevel);
  }

  /**
   * Retrieves min log level from integration database.
   *
   * @return int
   *   CleverReach Log level.
   */
  public function getMinLogLevel() {
    return $this->get('min_log_level');
  }

  /**
   * Retrieves access token from integration database.
   *
   * @return string
   *   CleverReach user access token.
   */
  public function getAccessToken() {
    if (empty($this->accessToken)) {
      $this->accessToken = $this->get('access_token');
    }

    return $this->accessToken;
  }

  /**
   * Save access token in integration database.
   *
   * @param string $accessToken
   *   CleverReach user access token.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setAccessToken($accessToken) {
    $this->set('access_token', $accessToken);
    $this->accessToken = $accessToken;
  }

  /**
   * Return whether article/product search is enabled or not.
   *
   * @return bool
   *   For this integration is enabled always.
   */
  public function isProductSearchEnabled() {
    return TRUE;
  }

  /**
   * Gets site name defined in configuration.
   *
   * @return string
   *   Site name.
   */
  public function getSiteName() {
    return variable_get('site_name', self::INTEGRATION_NAME);
  }

  /**
   * Retrieves parameters needed for article search registrations.
   *
   * @return array
   *   Keyed array, name, url, password.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function getProductSearchParameters() {
    $articleSearchUrl = url('cleverreach/search', array('absolute' => TRUE));

    return array(
      'url' => $articleSearchUrl,
      'name' => $this->getIntegrationListName(),
      'password' => $this->getArticleSearchEndpointPassword(),
    );
  }

  /**
   * Retrieves name of the CleverReach list (group).
   *
   * @return string
   *   CleverReach group name.
   */
  public function getIntegrationListName() {
    return "{$this->getIntegrationName()} - {$this->getSiteName()}";
  }

  /**
   * Retrieves integration name.
   *
   * @return string
   *   Integration name.
   */
  public function getIntegrationName() {
    return self::INTEGRATION_NAME;
  }

  /**
   * Retrieves integration ID.
   *
   * @return int
   *   CleverReach Group ID / Integration ID.
   */
  public function getIntegrationId() {
    return $this->get('integration_id');
  }

  /**
   * Saves created groupID / listID in CleverReach to integration configuration.
   *
   * @param int $id
   *   CleverReach Group ID.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setIntegrationId($id) {
    $this->set('integration_id', $id);
  }

  /**
   * Retrieves user account ID.
   *
   * @return string
   *   CleverReach account ID.
   */
  public function getUserAccountId() {
    $userInfo = $this->getUserInfo();
    return empty($userInfo['id']) ? '' : $userInfo['id'];
  }

  /**
   * Set default logger status, enabled or disabled.
   *
   * @param bool $status
   *   Current logger status.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setDefaultLoggerEnabled($status) {
    $this->set('default_logger_status', $status);
  }

  /**
   * Return whether default logger is enabled or not.
   *
   * @return bool
   *   If enabled returns true, otherwise false.
   */
  public function isDefaultLoggerEnabled() {
    return $this->get('default_logger_status') === TRUE;
  }

  /**
   * Gets the number of maximum allowed started task at the point in time.
   *
   * This number will determine how many tasks can be in "in_progress" status at
   * the same time.
   *
   * @return int
   *   Number of maximum allowed started tasks.
   */
  public function getMaxStartedTasksLimit() {
    return $this->get('max_started_task_limit');
  }

  /**
   * Sets the number of maximum allowed started task at the point in time.
   *
   * @param int $maxStartedTaskLimit
   *   Number of maximum allowed started tasks.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setMaxStartedTaskLimit($maxStartedTaskLimit) {
    $this->set('max_started_task_limit', $maxStartedTaskLimit);
  }

  /**
   * Get user information.
   *
   * @return array
   *   Keyed array with user info.
   */
  public function getUserInfo() {
    if (empty($this->userInfo)) {
      $this->userInfo = $this->get('user_info');
    }

    return $this->userInfo === NULL ? array() : $this->userInfo;
  }

  /**
   * Save user information in integration database.
   *
   * @param array $userInfo
   *   Keyed array with user info.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setUserInfo($userInfo) {
    $this->set('user_info', $userInfo);
    $this->userInfo = $userInfo;
  }

  /**
   * Automatic task runner wakeup delay in seconds.
   *
   * Task runner will sleep at the end of its lifecycle for this value seconds
   * before it sends wakeup signal for a new lifecycle. Return null to use
   * default system value (10).
   *
   * @return int|null
   *   Wakeup delay in seconds.
   */
  public function getTaskRunnerWakeupDelay() {
    return $this->get('task_runner_wakeup_delay');
  }

  /**
   * Set automatic task runner wakeup delay in seconds.
   *
   * Task runner will sleep at the end of its lifecycle for this value seconds
   * before it sends wakeup signal for a new lifecycle. Return null to use
   * default system value (10).
   *
   * @param int $taskRunnerWakeUpDelay
   *   Wakeup delay in seconds.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setTaskRunnerWakeUpDelay($taskRunnerWakeUpDelay) {
    $this->set('task_runner_wakeup_delay', $taskRunnerWakeUpDelay);
  }

  /**
   * Gets maximal time in seconds allowed for runner instance to stay alive.
   *
   * After this period system will automatically start new runner instance and
   * shutdown old one. Return null to use default system value (60).
   *
   * @return int|null
   *   Max execution time of runner in seconds.
   */
  public function getTaskRunnerMaxAliveTime() {
    return $this->get('max_alive_time');
  }

  /**
   * Set maximal time in seconds allowed for runner instance to stay in alive.
   *
   * @param int $taskRunnerMaxAliveTime
   *   Max execution time of runner in seconds.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setTaskRunnerMaxAliveTime($taskRunnerMaxAliveTime) {
    $this->set('max_alive_time', $taskRunnerMaxAliveTime);
  }

  /**
   * Gets maximum number of failed task execution retries.
   *
   * System will retry task execution in case of error until this number is
   * reached. Return null to use default system value (5).
   *
   * @return int
   *   Number of task retries before it is considered as failed.
   */
  public function getMaxTaskExecutionRetries() {
    return $this->get('max_task_execution_retries');
  }

  /**
   * Sets maximum number of failed task execution retries.
   *
   * @param int $maxTaskExecutionRetries
   *   Number of task retries before it is considered as failed.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setMaxTaskExecutionRetries($maxTaskExecutionRetries) {
    $this->set('max_task_execution_retries', $maxTaskExecutionRetries);
  }

  /**
   * Gets max inactivity period for a task in seconds.
   *
   * After inactivity period is passed, system will fail such tasks as expired.
   * Return null to use default system value (30).
   *
   * @return int
   *   Max inactivity period or runner in seconds.
   */
  public function getMaxTaskInactivityPeriod() {
    return $this->get('max_task_inactivity_period');
  }

  /**
   * Sets max inactivity period for a task in seconds.
   *
   * @param int $maxTaskInactivityPeriod
   *   Max inactivity period or runner in seconds.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setMaxTaskInactivityPeriod($maxTaskInactivityPeriod) {
    $this->set('max_task_inactivity_period', $maxTaskInactivityPeriod);
  }

  /**
   * Gets task runner status information.
   *
   * @return array
   *   Key array containing guid and timestamp.
   */
  public function getTaskRunnerStatus() {
    $runnerStatus = $this->get('task_runner_status');
    return $runnerStatus === NULL ? array() : $runnerStatus;
  }

  /**
   * Sets task runner status information as array.
   *
   * @param string $guid
   *   Unique string.
   * @param int $timestamp
   *   Unix timestamp.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setTaskRunnerStatus($guid, $timestamp) {
    $this->set(
        'task_runner_status',
        array(
          'guid' => $guid,
          'timestamp' => $timestamp,
        )
    );
  }

  /**
   * Gets batch size for synchronization set in configuration.
   *
   * @return int
   *   Batch size.
   */
  public function getRecipientsSynchronizationBatchSize() {
    return $this->get('recipient_sync_batch_size');
  }

  /**
   * Sets synchronization batch size.
   *
   * @param int $batchSize
   *   Batch size.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setRecipientsSynchronizationBatchSize($batchSize) {
    $this->set('recipient_sync_batch_size', $batchSize);
  }

  /**
   * Gets CleverReach client id for Drupal integration.
   *
   * @return string
   *   Unique client id generated by CleverReach.
   */
  public function getClientId() {
    return self::CLIENT_ID;
  }

  /**
   * Gets CleverReach client secret for Drupal integration.
   *
   * @return string
   *   Unique client secret generated by CleverReach.
   */
  public function getClientSecret() {
    return self::CLIENT_SECRET;
  }

  /**
   * Gets task execution context.
   *
   * @return string
   *   Context in which task is being executed. If no context is
   *   provided empty string is returned (global context)
   */
  public function getContext() {
    return $this->context;
  }

  /**
   * Sets task execution context.
   *
   * When integration supports multiple accounts (middleware integration)
   * proper context must be set based on middleware account that is using
   * core library functionality. This context should then be used by
   * business services to fetch account specific data.Core will set context
   * provided upon task enqueueing before task execution.
   *
   * @param string $context
   *   Context to set.
   */
  public function setContext($context) {
    $this->context = $context;
  }

  /**
   * Return true if recipient is active by default, otherwise false.
   *
   * @return bool
   *   Default subscription behavior.
   */
  public function getDefaultRecipientStatus() {
    return $this->get('default_recipient_status') === TRUE;
  }

  /**
   * Sets default recipient status.
   *
   * @param bool $status
   *   Default subscription behavior.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setDefaultRecipientStatus($status) {
    $this->set('default_recipient_status', $status);
  }

  /**
   * Checks if initial sync is configured.
   *
   * @return bool
   *   If initial sync is configured, returns true, otherwise false.
   */
  public function isConfiguredInitialSync() {
    return $this->get('configured_initial_sync') === TRUE;
  }

  /**
   * Sets flag that indicate whether initial sync is configured or not.
   *
   * @param bool $value
   *   Indicator whether initial sync is configured or not.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setConfiguredInitialSync($value = TRUE) {
    $this->set('configured_initial_sync', $value);
  }

  /**
   * Return whether first email is already built or not.
   *
   * @return bool
   *   When first email is built, return true, otherwise false.
   */
  public function isFirstEmailBuilt() {
    return $this->get('first_email_built') === TRUE;
  }

  /**
   * Sets information whether first email is built.
   *
   * @param bool $value
   *   Indicator whether first email is build or not.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setIsFirstEmailBuilt($value) {
    $this->set('first_email_built', $value);
  }

  /**
   * Gets article search endpoint password.
   *
   * @return string
   *   Unique article search password stored in database
   *   when module is installed.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function getArticleSearchEndpointPassword() {
    return $this->get('article_search_password');
  }

  /**
   * Sets article search endpoint password.
   *
   * @param string $password
   *   Random generated string.
   *
   * @throws Exception
   *   When module is not installed.
   */
  public function setArticleSearchEndpointPassword($password) {
    $this->set('article_search_password', $password);
  }

  /**
   * Gets queue name used for this integration.
   *
   * @return string
   *   Queue name used for integration.
   */
  public function getQueueName() {
    return 'drupalDefault';
  }

  /**
   * Sets key-value pair to Drupal configuration table.
   *
   * @param string $key
   *   Identifier to store value in configuration.
   * @param mixed $value
   *   Value to associate with identifier.
   *
   * @throws Exception
   */
  private function set($key, $value) {
    if (!$this->get('installed')) {
      // Kill all processes as soon as module uninstall is detected.
      throw new Exception(
        'CleverReach module is not currently installed, abort all processes.'
      );
    }

    variable_set("cleverreach_$key", $value);
  }

  /**
   * Gets configuration by key from Drupal configuration table.
   *
   * @param string $key
   *   Identifier to store value in configuration.
   *
   * @return mixed|null
   *   Configuration value.
   */
  private function get($key) {
    return variable_get("cleverreach_$key");
  }

}
